"""Codegen package responsive for generation of code for assembling.

Provides different backends for OSxARCH pair
"""

from .generator import generate_code_for_assembler

__all__ = ["generate_code_for_assembler"]
